from typing import Optional

from django import forms

from .layout import _Section


class JSONSchema:
    """Encapsulation of a JSON form schema."""

    def __init__(self):
        self.schema = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "type": "object",
            "properties": {},
        }

    @staticmethod
    def generate_field(field):
        """Generate schema for one form field."""
        # Defaults for a generic field
        new_field = {
            "type": "string",
            "title": str(field.label or ""),
            "description": str(field.help_text or ""),
            "readOnly": field.disabled,
            "required": field.required,
        }

        # Inspect field and update schema accordingly
        if isinstance(field.widget, forms.TextInput):
            new_field["type"] = "string"
        elif isinstance(field.widget, forms.NumberInput) and isinstance(field, forms.IntegerField):
            new_field["type"] = "integer"
        elif isinstance(field.widget, forms.NumberInput):
            new_field["type"] = "number"
        elif isinstance(field.widget, forms.EmailInput):
            new_field["type"] = "string"
            new_field["format"] = "email"
        elif isinstance(field.widget, forms.URLInput):
            new_field["type"] = "string"
            new_field["format"] = "url"
        elif isinstance(field.widget, forms.PasswordInput):
            new_field["type"] = "string"
            new_field["format"] = "password"
            new_field["x-display"] = "password"
        elif isinstance(field.widget, forms.HiddenInput):
            new_field["type"] = "string"
            new_field["format"] = "hidden"
        elif isinstance(field.widget, forms.FileInput):
            new_field["type"] = "string"
            new_field["contentMediaType"] = "image/*" if isinstance(field, forms.ImageField) else "*"
            new_field["writeOnly"] = True
        elif isinstance(field.widget, forms.Textarea):
            new_field["x-display"] = "textarea"
        elif isinstance(field.widget, forms.DateInput):
            new_field["type"] = "string"
            new_field["format"] = "date"
        elif isinstance(field.widget, forms.DateTimeInput):
            new_field["type"] = "string"
            new_field["format"] = "date-time"
        elif isinstance(field.widget, forms.TimeInput):
            new_field["type"] = "string"
            new_field["format"] = "time"
        elif isinstance(field.widget, forms.CheckboxInput):
            new_field["type"] = "boolean"
        elif type(field.widget) in [forms.Select, forms.SelectMultiple, forms.RadioSelect, forms.CheckboxSelectMultiple,
                                    forms.NullBooleanSelect]:
            one_of = []
            for const, title in field.widget.choices:
                one_of.append(dict(const=str(const), title=str(title)))

            if field.widget.allow_multiple_selected:
                new_field["type"] = "array"
                new_field["items"] = {
                    "type": "string",
                    "oneOf": one_of
                }
            else:
                new_field["type"] = "string"
                new_field["oneOf"] = one_of

            if isinstance(field.widget, forms.RadioSelect):
                new_field["x-display"] = "radio"
            elif isinstance(field.widget, forms.CheckboxSelectMultiple):
                new_field["x-display"] = "checkbox"
        else:
            raise TypeError(f"Unsupported field/widget: {repr(field)}, {repr(field.widget)}")

        return new_field

    def add_field(self, name, field, metadata: Optional[dict] = None):
        """Add a form field to this schema."""
        new_field = self.generate_field(field)
        if metadata and metadata.get("section_name") and self.schema["properties"].get(
                metadata["section_name"]
        ) is not None:
            self.schema["properties"][metadata["section_name"]]["properties"] = \
                self.schema["properties"][metadata["section_name"]]["properties"] or {}
            self.schema["properties"][metadata["section_name"]]["properties"][name] = new_field
        else:
            self.schema["properties"][name] = new_field

    @staticmethod
    def generate_section(section: _Section):
        """Generate schema for a form section."""
        sec = {
            "title": section.title,
            "type": "object",
            "properties": {}
        }

        if section.description:
            sec["description"] = section.description

        return sec

    def add_section(self, section: _Section):
        """Add a form section to this schema."""
        self.schema["properties"][section.codename] = self.generate_section(section)

    def update_properties(self, props: dict):
        """Merge new property values into schema properties."""
        self.schema["properties"].update(props)
