"""Re-implementation of django-material's form layout mechanics.

This implementation provides a drop-in replacement for django-material's
Layout, Row, and Fieldset classes, which allow rendering into a JSON schema.

cf. http://docs.viewflow.io/material_forms.html
"""

from abc import ABC
from dataclasses import dataclass
from typing import Optional, Union

from django.utils.text import slugify


class LayoutNode(ABC):
    """Abstract node in a form layout."""
    def __init__(self, *elements):
        self.elements = elements

    def build_schema(self, schema, form_fields) -> dict:
        """Render this node into a JSON schema fragment."""
        props = {}
        for field in self.elements:
            if isinstance(field, LayoutNode):
                # Recurse and add fragment of the sub-node
                built_schema = field.build_schema(schema, form_fields)
                props.update(built_schema["properties"])
            else:
                # Add verbatim field
                props[field] = schema.generate_field(form_fields[field])

        return {"properties": props}


@dataclass
class _Section:
    """Visual section in a form."""
    codename: str
    title: str
    description: Optional[str] = None


class Row(LayoutNode):
    """Visual row in a form."""

    def build_schema(self, schema, form_fields):
        """Render this row as a JSON schema fragment."""
        row_name = f"row-{id(self)}"
        fields = super().build_schema(schema, form_fields)["properties"]
        for k in fields.keys():
            # Make row responsive by extending each field to full width on smaller viewports
            fields[k].setdefault("x-options", {}).setdefault("fieldColProps", {}).update({"cols": 12, "md": ""})

        return {
            "properties": {
                row_name: {
                    "x-cols": 12,
                    "properties": fields
                }
            }
        }


class Fieldset(LayoutNode):
    """Visual set of fields in a form."""
    def __init__(self, name: Union[tuple[str, str], str], *elements):
        super().__init__(*elements)
        codename = str(slugify(name))

        if isinstance(name, tuple):
            self.section = _Section(codename, *name)
        else:
            self.section = _Section(codename, name)

    def build_schema(self, schema, form_fields) -> dict:
        """Render this fieldset as a JSON schema fragment."""
        section = schema.generate_section(self.section)
        section["properties"].update(super().build_schema(schema, form_fields)["properties"])
        return {
            "properties": {
                self.section.codename: section
            }
        }


class Layout(LayoutNode):
    """Full form layout as the root node."""
